import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.datasets as dsets
import general_energy_nets.predictive_coding as pc
import torchvision
import torch.optim as optim
import argparse
import matplotlib
import matplotlib.pyplot as plt
import torchvision.transforms as transforms
from tqdm import tqdm
from torchvision import datasets
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torchvision.utils import save_image
from torch.autograd import Variable
import Utils as u
import statistics
import warnings
import sys
import numpy as np
import gc

if not sys.warnoptions:
    warnings.simplefilter("ignore")

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')



# Hyperparameters of the model
input_dim = 3*32*32
output_dim = 10

# Number of neurons
N = 7000
#1000 - 89.34 ± 0.08
#2500 - 81.82 ± 1.47

# Epochs needed to train the model
epochs = 1

# Number of iterations performed by the train and test inference on every batch of data.
num_iterations = 10000
iterations_test = 10000
t_test = 250

# Parameters of the algorithm. 
p_lr = 0.00005
x_lr = 0.5
xlr_test = 0.1
xl_discount = 0.9

# Name of the dataset
dataset_name = 'CIFAR10'

s = 0.1
wd = 1
acti_func = nn.ReLU()
artial = 50
batch_size = 100 #just for training
test_batch_size = 2000

if 'MNIST' in dataset_name:
    fraction = 392
    input_shape = (-1,1,28,28)
if 'CIFAR' in dataset_name:
    fraction = 512
    input_shape = (-1,3,32,32)

#################################################################################
def generation(pc_predictor,output_dim,s):

    def callback_generation(t, pc_layer, output_dim, pc_predictor):
        model[0].weight.data.copy_(model[0].weight.data * M.cuda())
        classes = torch.eye(output_dim).cuda()
        pc_layer.get_x().data[:,-output_dim:].copy_(classes)

        if t == 0:
            pc_predictor.inputs = pc_layer.get_x()

    callback_generation_kwargs={
        'pc_layer': pc_layer,
        'output_dim': output_dim,
        'pc_predictor': pc_predictor,
    }

    fake_inputs = torch.rand(output_dim,N).cuda()
    pc_predictor.train_on_batch(inputs=fake_inputs,is_log_progress=False,callback_after_t=callback_generation,callback_after_t_kwargs=callback_generation_kwargs)
    generated_images = pc_layer.get_x().data[:, :input_dim].view(output_dim,1,32,32)[:10]

    save_image(generated_images.cpu(), f"./assembly_results/generated_images/{dataset_name}: Epoch={epoch}-N={N},wd={wd},xlr={x_lr},plr={p_lr},T={num_iterations},mask_p={s}.png", nrow=10)

#########################################################
def accuracy(pc_predictor,dataloader,s):

    total = 0.0
    correct = 0.0
    pred_matrix_per_digit=np.zeros((10,10))

    def callback_acc(t, pc_layer, data, pc_predictor):
        model[0].weight.data.copy_(model[0].weight.data * M.cuda()) #used to add sparsity
        pc_layer.get_x().data[:,:input_dim].copy_(data.view(-1,input_dim))

        if t == 0:
            pc_predictor.inputs = pc_layer.get_x() #input?
    print(len(dataloader))

    for data, labels in dataloader:

        callback_acc_kwargs={
            'pc_layer': pc_layer,
            'data': data,
            'pc_predictor': pc_predictor,
        }

        fake_inputs = torch.rand(test_batch_size,N).cuda()
        pc_predictor.train_on_batch(inputs=fake_inputs,is_log_progress=False,callback_after_t=callback_acc,callback_after_t_kwargs=callback_acc_kwargs)
        outputs = pc_layer.get_x().data[:, -output_dim:]
        _, predicted = torch.max(outputs.data, 1)
        total += labels.size(0)
        correct += (predicted.cpu() == labels.cpu()).sum()
        pred_matrix_per_digit +=accuracy_for_each_digit(predicted,labels)

    
    return int(10000 * correct / total)/100, pred_matrix_per_digit

def accuracy_for_each_digit(predicted, labels):
    times = np.zeros((10,10))
    for i in range(0, 10):
        mask = (labels.cpu()==i)
        data_labeled_i = torch.masked_select(predicted,mask)
        for j in range(0,10):
            mask2 = data_labeled_i.cpu()==j
            times[i][j] = int(mask2.sum())
    return times

def reconstruct(model, dataloader, dataset_name,M):

    def prediction(t, pc_layer,pc_trainer):
        model[0].weight.data.copy_(model[0].weight.data * M.cuda())
        if 'CIFAR' in dataset_name:
            pc_layer.get_x().data[:,:fraction].copy_(lab_b)
            pc_layer.get_x().data[:,1024:1024+fraction].copy_(lab_r)
            pc_layer.get_x().data[:,2048:2048+fraction].copy_(lab_g)
        if 'MNIST' in dataset_name:
            pc_layer.get_x().data[:,:fraction].copy_(lab)
        if t == 0:
            prediction_trainer.inputs = pc_layer.get_x()
        
    callback_after_t_kwargs={'pc_layer':pc_layer.cuda(),'pc_trainer':pc_trainer}

    loss = 0.0
    for i, (images, labels) in enumerate(dataloader):
        model.train()
        num_rows = 10

        if 'CIFAR' in dataset_name:
            lab_b = images[:,:fraction]
            lab_r = images[:,1024:1024+fraction]
            lab_g = images[:,2048:2048+fraction]
        if 'MNIST' in dataset_name:
            lab = images[:,:fraction]

        if 'CIFAR' in dataset_name:
            zeros = torch.zeros((batch_size,3072))
            zeros[:,:fraction] = lab_b
            zeros[:,1024:1024+fraction] = lab_r
            zeros[:,2048:2048+fraction] = lab_g
        if 'MNIST' in dataset_name:
            zeros = torch.zeros((batch_size,784))
            zeros[:,:fraction] = lab

        images = Variable(images.view(-1,1).cuda())
        in_data=torch.zeros(batch_size, N).cuda()
        results = prediction_trainer.train_on_batch(in_data.float(),is_log_progress=False,callback_after_t=prediction,callback_after_t_kwargs=callback_after_t_kwargs)
        loss += results['overall'][t_test-1]
        reconstruction = pc_layer.get_x().data.cpu().view(batch_size,-1)[:,:input_dim]
        noise = images.view(batch_size,-1).cpu() - reconstruction
        print(noise.shape)
        bar1_list = [0.01, 0.005]
        bar2_list = [0.1]
        grid2 = [(ll, kk) for ll in bar1_list for kk in bar2_list]
        for bar1, bar2 in grid2:
            mse = torch.mean(noise**2, 1, True)

            n_correct_recon = int(torch.sum(mse<bar1))
            avg_correct_pixels = int(torch.mean(torch.sum((abs(noise)<bar2).type(torch.DoubleTensor), 1,True))) - fraction*3
            print(mse.shape)
            print(n_correct_recon)
            print(avg_correct_pixels)
            print(f"N={N},s={s},wd={wd},partial={partial},frac={fraction},T={num_iterations},xlr={x_lr},plr={p_lr},T_Test={t_test}",file=open('mse_cifar_0827.txt','a'))
            print(f"avg correct recon pixels in each image = {avg_correct_pixels}, acc = {int(10000*avg_correct_pixels/((1024-fraction)*3))/100}% where when diff(orginal, recon) < {bar2} considered as correctness",file=open('mse_cifar_0827.txt','a'))
            print(f"number of correct recon images: {n_correct_recon}, acc={int(10000*n_correct_recon/(partial*10))/100}% where mse of a recon < {bar1} considered as a corecct recon\n",file=open('mse_cifar_0827.txt','a'))
        print("done")
        
        if partial > 0: 
            image = torch.cat((zeros.view(input_shape)[0::partial], 
                                reconstruction.view(input_shape)[0::partial],
                                images.view(input_shape)[0::partial].cpu(),
                                noise.view(input_shape)[0::partial]))
            save_image(image.cpu(), f"./assembly_results/reconstruction0827_ns_cifar/-{dataset_name} - N={N},s={s},partial={partial},frac={fraction},T={num_iterations},xlr={x_lr},plr={p_lr},T_Test={t_test}.png", nrow=10)
        else:
            image = torch.cat((zeros.view(input_shape)[0:10],
                                reconstruction.view(input_shape)[0:10],
                                images.view(input_shape)[0:10].cpu(),
                                noise.view(input_shape)[0:10]))
            save_image(image.cpu(), f"./assembly_results/reconstruction0827_ns_cifar/-{dataset_name} - N={N},s={s},batch={batch_size},frac={fraction},T={num_iterations},xlr={x_lr},plr={p_lr},T_Test={t_test},Epoch={epoch}.png", nrow=10)
        break

###########################################################################################
best_accuracies = []
pred_mat=np.zeros((10,10))

N_list = [8000]#[11000, 12000, 13000]#[4000, 5000, 6000, 7000, 8000, 9000, 10000]#[24000,27000]8000
s_list = [0.8]#[ 0.1, 0.3, 0.5, 0.8, 1]#[0.8, 1] #
partial_list = [3]#[3, 5, 10, 25, 50]# 1,2,4 # 2, 3, ,2,3, 
frac_list = [512]#[512,int(1024/3),int(1024/4),int(1024/5),int(1024/6)]#  512,, ,256,204
grid = [(x, y, z,q) for x in N_list for y in s_list for z in partial_list for q in frac_list]

for N, s, partial, fraction in grid:
    batch_size=partial*10
    train_loader, test_loader = u.load_dataset(dataset_name, partial, batch_size, test_batch_size, augmentation = False, shuffle = True)
    print('N={},s={},partial={},batch_size={},fraction={}'.format(N, s, partial, batch_size,fraction))
    for seed in range(10):
        if s < 1:
            M = (torch.rand(size=(N,N)) < s).int()
        else:
            M = torch.ones((N,N))

        M[:input_dim,:input_dim] = torch.zeros((input_dim,input_dim))
        M[-output_dim:,-output_dim:] = torch.zeros((output_dim,output_dim))
        M.cuda()
        pc_layer = pc.PCLayer()
        model = nn.Sequential(nn.Linear(N, N),acti_func,pc_layer)

        model.train()
        model.cuda()

        pc_trainer = pc.PCTrainer(model,optimizer_x_kwargs={'lr': x_lr},optimizer_p_fn=optim.Adam, optimizer_p_kwargs={'lr': p_lr,'weight_decay':wd}, T = num_iterations,x_lr_discount=xl_discount,plot_progress_at=[],early_stop_condition='True')
        pc_predictor = pc.PCTrainer(model, optimizer_x_kwargs={'lr': xlr_test}, optimizer_p_kwargs={'lr': 0},optimizer_p_fn=optim.SGD, T = iterations_test,x_lr_discount=xl_discount,plot_progress_at=[],early_stop_condition='True')

        def callback_after_t(t, pc_layer, data, target, pc_trainer):
            model[0].weight.data.copy_(model[0].weight.data * M.cuda())
            pc_layer.get_x().data[:, :input_dim].copy_(data)
            pc_layer.get_x().data[:, -output_dim:].copy_(target)
            if t == 0:
                pc_trainer.inputs = pc_layer.get_x()

        acc = []
        pred_mat_per_epoch=[]
        for epoch in range(epochs):
            print("This is epoch ", epoch)
            print("Training starts...")
            for i, (data, target) in enumerate(train_loader):
                print(i)
                in_data=torch.zeros(batch_size, N).cuda()
                callback_after_t_kwargs={'pc_layer':pc_layer.cuda(),'data':data.cuda(),'target':target.cuda(),'pc_trainer':pc_trainer}
                pc_trainer.train_on_batch(inputs=in_data, is_log_progress=False,callback_after_t=callback_after_t,callback_after_t_kwargs=callback_after_t_kwargs)
            print("Prediction starts...")
            if epoch % 1 == 0:
                prediction_trainer = pc.PCTrainer(model, optimizer_x_kwargs={'lr': xlr_test}, optimizer_p_kwargs={'lr': 0},optimizer_p_fn=optim.SGD,x_lr_discount=0.9, T = t_test,early_stop_condition='True')

                generation(pc_predictor,output_dim,s)
                reconstruct(model, train_loader,dataset_name,M)
                acc_overall,pred_matrix_per_digit = accuracy(pc_predictor,test_loader,s)
                acc.append(acc_overall)
                pred_mat_per_epoch.append(pred_matrix_per_digit)

        best_accuracies.append(max(acc))
        for ind in pred_mat_per_epoch:
            pred_mat+= ind

        del pc_trainer
        del prediction_trainer
        del pc_predictor
        gc.collect()
        torch.cuda.empty_cache()

    avg_acc = int(100*sum(best_accuracies)/len(best_accuracies))/100
    std = int(100*statistics.stdev(best_accuracies))/100
    acc_per_digit={}
    most_common_err={}
    total_pred_mat = np.sum(pred_mat,axis=1)
    for ind in range(0,10):
        acc_per_digit[ind]=int(10000*pred_mat[ind][ind]/total_pred_mat[ind])/100.0
        max_error = 0
        max_index = 0
        for k in range(0,10):
            if k!=ind:
                if pred_mat[ind][k]>max_error:
                    max_error=pred_mat[ind][k]
                    max_index = k
        most_common_err[ind]=max_index
    print('{} - partial: {}, N:{}, wd={}, batch={},T={},s={},xl_discount={} - Acc = {} ± {}'.format(dataset_name,partial,N,wd,batch_size,num_iterations,s,xl_discount,avg_acc,std),file=open('cifar_rough_result.txt','a'))
    print('Acc for each digit: {}'.format(acc_per_digit),file=open('cifar_rough_result.txt','a'))
    print('Most common error of each digit: {}'.format(most_common_err),file=open('cifar_rough_result.txt','a'))